from AccessControl import ClassSecurityInfo

from interfaces import IContextCriteriaMatcher

from Products.ZCatalog.ZCatalog import ZCatalog
from Products.CMFPlone import CatalogTool

from Acquisition import aq_base
from AccessControl import getSecurityManager
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.Expression import Expression
from Products.CMFCore.Expression import createExprContext

from Acquisition import aq_base, aq_parent, aq_inner


import types

from config import *

class CriterionWrapper:
    '''Decorates a criterion for cataloging'''

    def __init__(self, criterion):
        self._criterion = criterion

    def __getattr__(self, name):
        return getattr(self._criterion, name)

    def getCategories(self):
        v = self._criterion.getCategories()
        if not v:
            return (MATCHALL,)
        return v

    def getContentTypes(self):
        v = self._criterion.getContentTypes()
        if not v:
            return (MATCHALL,)
        return v

    def getMatchPath(self):
        v = self._criterion.getMatchPath()
        if not v:
            return MATCHALL # fake path
        return v

    def getStates(self):
        v = self._criterion.getStates()
        if not v:
            return (MATCHALL,)
        return v

    def getTemplateIds(self):
        v = self._criterion.getTemplateIds()
        if not v:
            return (MATCHALL,)
        return v

    def getMatchRoles(self):
        v = self._criterion.getMatchRoles()
        if not v:
            return (MATCHALL,)
        return v


class ContextCriteriaMatcher(ZCatalog):
    """
    Base class for tools managing objects derived from
    ContextCriteriaMixin. Derives from ZCatalog to provide a
    catalogue for such objects. The method getItemsInContext () will
    query the catalogue for the objects matching the current type.
    """
    
    security = ClassSecurityInfo()

    __implements__ = (ZCatalog.__implements__, IContextCriteriaMatcher,)

    # tool-constructors have no id argument, the id is fixed
    def __init__(self):
        ZCatalog.__init__(self, self.getId())
        self._initIndexes()

    security.declarePublic('match')
    def match(self, context, **kw):
        """Return a catalog query result set with matches for the given context
        and any additional search parameters given as keyword arguments. If
        a returned item has a TALES condition, evaluate this and only return
        those which evaluate to true. Only returns active items.
        """
        results = self.matchByContext(context, **kw)
        nr = []
        for brain in results:
            if self.evaluateCondition(brain, context):
                nr.append(brain)
        return nr
        
    # Helper methods

    # We don't access these methods directly TTW - they are called by
    # getActions() in the actions tool, for example, which provides a better-
    # defined UI.
    
    security.declarePrivate('matchByContext')
    def matchByContext(self, context, **kw):
        """Get a catalog search result of all active items matching the
        current context. Criteria not directly obtainable from the contex may
        be passed in as keyword arguments.
        """
        
        query = {}
        
        if 'match_content_types' not in kw:
            query ['match_content_types'] = context.getTypeInfo().getId()
          
        if 'match_path' not in kw:
            query ['match_path'] = '/'.join(context.getPhysicalPath()) 
        
        if 'match_user_roles' not in kw:
            query ['match_user_roles']        = \
                getSecurityManager().getUser().getRolesInContext(context)
                
        if 'match_workflows_states' not in kw:
            query ['match_workflow_states']    = \
                getToolByName(self, 'portal_workflow').getInfoFor(context,
                                                    'review_state', None)
        
        # match_categories and match_template_ids must be passed in explicitly.
        # This is done in getActions() in ContextActionsTool.
        
        # Any keyword arguments override what we obtain from the context
        query.update(kw)
        
        # Ensure we only get active actions
        query['active'] = 1
        
        # Allow subclasses to filter the query
        self.filterContextQuery(query)
        
        return self.searchResults(None, **query)

    security.declarePrivate('filterContextQuery')
    def filterContextQuery(self, query):
        """Subclasses mixing in ContextCriteriaMatcher may implement this method
        to filter the query being sent to searchResults(). The 'query' param is
        a dict containing catalog indexes and keys and search criteria as 
        values. Hence, subclasses can modify this dict as appropriate to do any
        final filtering before the query is executed.
        """
        pass

    # searchResults has inherited security assertions.
    def searchResults(self, REQUEST=None, **kw):
        """Calls ZCatalog.searchResults with the extra MATCHALL wrapper, to
        ensure that if an item does not have a value set for a given criterion,
        it is counted as a match, not a mismatch.
        """
        def matchall(v):
            if type(v) in (types.ListType, types.TupleType):
                return list(v) + [MATCHALL]
            else:
                return [v, MATCHALL]

        for criterion, v in kw.items():
            if criterion.startswith('match_'):
                if type(v) == types.DictType:
                    v['query'] = matchall(v['query'])
                else:
                    kw[criterion] = matchall(v)

        return apply(ZCatalog.searchResults, (self, REQUEST), kw)

    # catalog_object has inherited security assertions
    def catalog_object(self, object, uid, idxs=[]):
        """Catalogs a decorated criterion object, automagically adding the 
        MATCHALL keyword wrapper.
        """
        return ZCatalog.catalog_object(self, CriterionWrapper(object), uid, idxs)

    # uniqueValuesFor has inherited security assertions
    def uniqueValuesFor(self, indexName):
        """Get a tuple of all unique values for the given index."""
        
        # This override is necessary because we don't want to return MATCHALL's
        # in uniqueValuesFor() - MATCHALL is an internal implementation detail.
        
        values = list(ZCatalog.uniqueValuesFor(self, indexName))
        
        if MATCHALL in values:
            values.remove(MATCHALL)
            
        return tuple(values)

    security.declarePrivate('evaluateCondition')
    def evaluateCondition(self, brain, context):
        """Evaluate an object's TALES condition to see if it should be 
        displayed
        """
        try:
            if brain.getExpression and context is not None:
                portal = getToolByName(self, 'portal_url').getPortalObject()
                
                # Find folder (code courtesy of CMFCore.ActionsTool)
                if context is None or not hasattr(context, 'aq_base'):
                    folder = portal
                else:
                    folder = context
                    # Search up the containment hierarchy until we find an
                    # object that claims it's a folder.
                    while folder is not None:
                        if getattr(aq_base(folder), 'isPrincipiaFolderish', 0):
                            # found it.
                            break
                        else:
                            folder = aq_parent(aq_inner(folder))
                
                __traceback_info__ = (folder, portal, context, 
                                        brain.getExpression)
                ec = createExprContext(folder, portal, context)
                return Expression(brain.getExpression)(ec)
            else:
                return 1
        except AttributeError:
            return 1

    security.declarePrivate('enumerateIndexes')
    def enumerateIndexes( self ):
        """Return a list of ( index_name, type, indexed attribute/method ) pairs 
        for the initial index set.
        """
        return ( ('Title',          'TextIndex' , None)
               , ('Type',           'FieldIndex', None)
               , ('created',        'FieldIndex', None)
               , ('modified',       'FieldIndex', None)
               , ('meta_type',      'FieldIndex', None)
               , ('getId',          'FieldIndex', None)
               , ('portal_type',    'FieldIndex', None)
               , ('install_package','FieldIndex', None)
               
               , ('match_content_types',    'KeywordIndex', 'getContentTypes')
               , ('match_path',             'ExtendedPathIndex', 'getMatchPath')
               , ('match_user_roles',       'KeywordIndex', 'getMatchRoles')
               , ('match_workflow_states',  'KeywordIndex', 'getStates')
               , ('match_categories',       'KeywordIndex', 'getCategories')
               , ('match_template_ids',     'KeywordIndex', 'getTemplateIds')
               
               # active is always searched
               , ('active', 'FieldIndex',   'isActive') 
               )
        # do we know how to add "indexed method to these?
    
    security.declarePrivate('enumerateColumns')
    def enumerateColumns(self):
        """Return a sequence of schema names to be cached as catalog metadata"""
        return ( 'Subject'
               , 'Title'
               , 'Description'
               , 'Type'
               , 'getIcon'
               , 'CreationDate'
               , 'getId'
               , 'portal_type'
               , 'getExpression'
               , 'getMatchPath'
               , 'getCategories'
               )

    security.declarePrivate('columns')
    def columns(self):
        """Avoid namespace clashes with Archetypes - make a new method to access 
        the metadata schema for the catalog in subclasses also deriving from 
        AT's baseObject, schema will be a Schema instance, not a method. 
        """
        return ZCatalog.schema(self)


    def _initIndexes(self):
        """Set up indexes and metadata, as enumared by enumerateIndexes() and
        enumerateColumns (). Subclasses can override these to inject additional
        indexes and columns.
        """
        base = aq_base(self)
        addIndex = self.addIndex
        addColumn = self.addColumn

        # Content indexes
        self._catalog.indexes.clear()
        for (index_name, index_type, extra) in self.enumerateIndexes():
            if extra is None:
               addIndex( index_name, index_type)
            else:
                p = Record(extra)
                # having to create the whole darn Record class, just to carry 
                # one argument must be the sickest interface invention i have 
                # seen till now. Someone should be shot for this  
                addIndex( index_name, index_type, extra=p )

        # Cached metadata
        self._catalog.names = ()
        self._catalog.schema.clear()
        for column_name in self.enumerateColumns():
            addColumn( column_name )

    security.declarePrivate('getAvailableCategories')
    def getAvailableCategories(self):
        """
        Get a list of the categories currently in use in this catalog. This
        helper method aids defining vocabularies for subclass tools which allow
        the user to select which categories to enable.
        """
        return self.uniqueValuesFor('match_categories')


class Record:
    """ a moron simple object for carrying the 'extra'-payload to index 
    constructors
    """
    
    def __init__(self, indexed_attrs):
        self.indexed_attrs = indexed_attrs

