# Utility methods for importing help pages from the filesystem
# during installation.

import os
import re

from Products.CMFCore.utils import getToolByName
from Globals import package_home

# Used to find title and description metadata for folders and bundles
TITLE_FILENAME="TITLE.txt"
DESCRIPTION_FILENAME="DESCRIPTION.txt"

def importHelpPages(self, productName, productGlobals, title, replace=False):
    """Import help pages from the filesystem in portal_contexthelp.
    
    Looks for a directory Products/${productName}/help. If this is found,
    a bundle with the name 'productName' will be created with the title 'title'.
    The product's globals() value must passed in as well, to enable the importer
    to find the correct 'help' directory. The standard way of achieving this
    is to set
    
        PRODUCT_GLOBALS = globals()
        
    in your product's config.py, and then pass the value of this in as the
    productGlobals parameter.
    
    Any folder within the top-level 'help' folder will be created inside 
    portal_contexthelp/${productName} as a ContextHelpFolder. Any text file 
    with a .hlp extension will be created as a ContextHelpPage. The format
    of a .hlp file is as follows:
    
        title: Help page title
        description: Short description of the help page
        categories: category1, category2
        contentTypes: Document, Image, AnotherTypeName
        path: /path/to/relevant/folder
        states: private, published, custom-state
        templateIds: template1, template2
        matchRoles: Owner, Manager
        expression: python:object.conditionToEvaluate()
        active: True
        related: ../parallel-folder/anotherpage.hlp, parallel-page.hlp
        Content-Type: text/structured

        Title!

          This is a test help page
          
          Look at this picture: "Screenshot":img:screenshot.gif

        Another paragraph here.
        
    The format must be exactly as shown, with the key names in the given order.
    Any field except 'title', 'categories' and 'active' may be left blank.
    Lists are comma-separated. Leading and trailing whitespace is trimmed.
    The Content-Type header must be followed by a blank line. Anything 
    thereafter will be parsed as the body text of the help page, with the given
    content type.
        
    Help pages may contain images, which are stored *inside* the help page.
    To add images to your help pages, place the image in the same folder as
    the .hlp page, and reference it from the body text of the page. The
    referenced image will be located and added to the help page.
    
    Help pages may also contain references to other help pages. In the portal,
    this is done using Archetypes UIDs, but these cannot be known in advance.
    Hence, use simple relative URLs for relating help pages. These will be
    translated to UIDs upon import.
    
    Help page bundles and folders may contain a title that is different form
    the folder name and a plain-text description. Since these cannot be set 
    directly on the filesystem, you must add files inside the folder or 
    bundle with the name 'TITLE.txt' and 'DESCRIPTION.txt' containing the 
    title and description, respectively. If TITLE.txt is not provided, 
    the title defaults to 
    
    If any bundle, folder or help page already exists, it will be silently
    ignored so long as 'replace' remains False. If 'replace' is True, the
    filesystem item will override the one in the portal.
    """

    tool = getToolByName(self, 'portal_contexthelp')
    rootDirectory = os.path.join(package_home(productGlobals), 'help')

    importer = HelpPageImport(tool, rootDirectory, productName, title, replace)
    importer.execute()
    
    
class HelpPageImporter:
    """Helper class which aids the import process.
    """
    
    def __init__(self, tool, rootDirectory, productName, title, replace=False):
        """Initialise the importer to import into the given tool instance, 
        from the given directory creating a bundle for the given product name.
        """
        self.tool          = tool
        self.rootDirectory = rootDirectory
        self.productName   = self.productName
        self.bundleTitle   = titleindexed_attrs
        self.replaceMode   = replace 
            
    def execute(self):
        """Perform the import. 
        """
        
        # Add initial bundle
        self.createBundle(self.rootDirectory, self.productName, 
                            self.bundleTitle)
        
        # Walk the directory tree
        for root, dirs, files in os.walk(directory):
            
            parent = self._getParentObject(root)
            
            for d in dirs:
                self.createFolder(root, d, parent)
                
            for f in files:
                if f.endswidth('.hlp'):
                    self.createHelpPage(root, f, parent)
        
    def createBundle(self, rootDirectory, productName, title):
        """Create/locate the product's bundle. Returns the bundle object.
        """
        
        description = self._getFolderDescription(rootDirectory)
        
        try:
            self.tool.addBundle(productName, title, description)
        except AttributeError:
            if self.replaceMode:
                bundle = self.tool.getBundle(productName)
                bundle.setTitle(title)
                bundle.setDescription(title)
        
    def createFolder(self, parentDirectory, directory, parent):
        """Create/locate the folder for the given directory, inside the
        given parent bundle/folder. Returns the folder object.
        """
        
        path = os.path.join(parentDirectory, directory)
        
        title = self._getFolderTitle(path)
        description = self._getFolderDescription(path)
        
        try:
            parent.addHelpFolder(directory, title, description)
        except AttributeError:
            if self.replaceMode:
                folder = parent.getHelpFolder(directory)
                folder.setTitle(title)
                folder.setDescription(title)
        
    def createHelpPage(self, parentDirectory, helpFile, parent):
        """Create/locate the help page in the given .hlp file (found in the
        given filesystem directory) inside the given parent bundle/folder.
        Returns the help page object.
        """
        
        path = os.path.join(parentDirectory, helpFile)
        fields = self._getHelpFileContents(path)
        
        page = None
        
        try:
            page = self.tool.addHelpPage(**fields)
        except AttributeError:
            if self.replaceMode:
                page = parent.getHelpPage(fields['id'])
                page.setTitle(fields['title'])
                page.setDescription(fields['description'])
                page.setBody(fields['body'], mimetype = fields['mimetype'])
                page.setCategories(fields['categories'])
                page.setContentTypes(fields['contentTypes'])
                page.setMatchPath(fields['path'])
                page.setStates(fields['states'])
                page.setTemplateIds(fields['templateIds'])
                page.setMatchRoles(fields['matchRoles'])
                page.setExpression(fields['expression'])
                page.setActive(fields['active'])
            else:
                return
                
        # XXX: What to do with references? Order of import becomes significant!
                
        # Add any images
        for image in self._getImageReferences(page):
            self._addImage(page, image)
        
    def _getParentObject(self, directory):
        """For the given directory name, find the corresponding bundle or
        folder object in the portal.
        """
        rootPathLength = len(self.rootDirectory)
        portalPath = directory[rootPathLength:]
        return self.tool.restrictedTraverse(portalPath)
    
    def _getFolderTitle(self, directory):
        """Get the contents of TITLE.txt for the given directory
        """
        try:
            titleFile = file(os.path.join(directory, TITLE_FILENAME), 'r')
            title = titleFile.read().strip()
            titleFile.close()
        except:
            return ''
        
    def _getFolderDescription(self, directory):
        """Get the contents of DESCRIPTION.txt for the given directory
        """
        try:
            titleFile = file(os.path.join(directory, DESCRIPTION_FILENAME), 'r')
            title = titleFile.read().strip()
            titleFile.close()
        except:
            return ''
        
    def _getHelpFileContents(self, path):
        """Parse the given help file. Return a dict containing keys 
        corresponding to the parameters to IContextHelpManager.addHelpPage().
        """
        pass
        
    def _getImageReferences(self, page):
        """Find any images referenced in the transformed-to-html body text
        of the given help page. Return a list of filenames for images which
        are referenced directly without a path - these need to be added
        inside the help page.
        """
        pass
        
    def _addImage(self, parentDirectory, page, image):
        """Load the image with the given name from the given directory, and add 
        it to the given page.
        """
        pass